import math
import pymunk
from pymunk import Vec2d
import gameobjects
from collections import defaultdict, deque
import images
import pygame

# NOTE: use only 'map0' during development!

MIN_ANGLE_DIF = math.radians(6) # input degrees

def angle_between_vectors(vec1, vec2):
    """ Since Vec2d operates in a cartesian coordinate space we have to
        convert the resulting vector to get the correct angle for our space.
    """
    vec = vec1 - vec2
    vec = vec.perpendicular()
    return vec.angle

def periodic_difference_of_angles(angle1, angle2):
    """ Return the periodic difference between two angles. """
    
    return (angle1 - angle2)%(2*math.pi)

class Ai:
    """ A simple ai that finds the shortest path to the target using
    a breadth first search. Also capable of shooting other tanks and or wooden
    boxes. """

    def __init__(self, tank,game_objects_list, tanks_list,space,currentmap):
        """ Ai constructor """
        self.tank               = tank
        self.game_objects_list  = game_objects_list
        self.tanks_list         = tanks_list
        self.space              = space
        self.currentmap         = currentmap
        self.flag = None
        self.MAX_X = currentmap.width - 1
        self.MAX_Y = currentmap.height - 1

        self.metal_path = False
        self.path = deque()
        self.move_cycle = self.move_cycle_gen()
        self.update_grid_pos()
        self.last_shot = pygame.time.get_ticks()

    def ray_cast(self):
        """Function for detecting ray cast for ai shoot function"""

        angle = self.tank.body.angle + math.pi/2 # adapt to cartesian angles
        x_bullet = self.tank.body.position[0]+0.5*math.cos(angle)
        y_bullet = self.tank.body.position[1]+0.5*math.sin(angle)
        ray_beg = (x_bullet, y_bullet) # same as bullet start
        ray_max1 = x_bullet + 100 * math.cos(angle)
        ray_max2 = y_bullet + 100 * math.sin(angle)
        ray_max = (ray_max1, ray_max2) # projectile line for bullet
        return self.space.segment_query_first(ray_beg, ray_max, 0, \
                                              pymunk.ShapeFilter())

    def update_grid_pos(self):
        """ This should only be called in the beginning,
            or at the end of a move_cycle. """
        
        self.grid_pos = self.get_tile_of_position(self.tank.body.position)

    def maybe_shoot(self):
        """ Makes a raycast query in front of the tank. If another tank
            or a wooden box is found, then we shoot. """
        
        res = self.ray_cast()
        if hasattr(res, "shape"):
            # check for the collision types
            if isinstance(res.shape.collision_type, int): 
                col_type = res.shape.collision_type
                if col_type == 1 or (col_type == 3 and \
                    res.shape.parent.destructable) or col_type == 2:
                    ai_bullet = self.tank.shoot()
                    if ai_bullet != None:
                        ##unfair Ai (times 1.3)
                        ai_bullet.body.velocity = ai_bullet.body.velocity*1.3 
                        self.game_objects_list.append(ai_bullet)


    def decide(self):
        """ Main decision function that gets called on every tick of the game.
        """
        self.maybe_shoot()
        next(self.move_cycle)    
    
    
    def find_shortest_path(self):
        """ A simple Breadth First Search using integer coordinates as our
            nodes. Edges are calculated as we go, using an external
            function. """
        
        source_node = self.grid_pos.int_tuple
        queue = deque() # neighbors (new nodes)
        visited = set()
        shortest_path = []
        queue.append((source_node, shortest_path))
        
        while queue: # all tiles in worst case...
            queue = self.astar_heu(queue) # rearrange according to heu!
            curr_tile, curr_shortest = queue.popleft()
            if self.get_target_tile() == curr_tile:
                curr_shortest.append(curr_tile)
                shortest_path = curr_shortest
                return deque(shortest_path[1:])
            neighbours = self.get_tile_neighbors(curr_tile)
            for neighbor in neighbours:
                neighbor = neighbor.int_tuple
                if neighbor not in visited:
                    queue.append((neighbor,curr_shortest + [curr_tile])) 
                    visited.add(neighbor)


    def astar_heu(self, input_queue):
        """ Heurstic function for calculating a path the
            flag using the shortest distance. """
        queue = deque(input_queue)
        astar_queue = deque()
        curr_min_dist = 1+Vec2d(self.tank.body.position\
                                ).get_distance(self.get_target_tile())
        while queue:
            curr_path =queue.popleft()
            temp_dist =Vec2d(curr_path[0]).get_distance(self.get_target_tile())
            if temp_dist < curr_min_dist:
                astar_queue.appendleft(curr_path)
                curr_min_dist = temp_dist
            else:
                astar_queue.append(curr_path)
        return astar_queue

    def move_cycle_gen(self):
        """ A generator that iteratively goes through all the required steps
            to move to our goal. """
        while True:
            path = self.find_shortest_path()
            if path == None:
                self.metal_path = True
                yield
                continue 
            if len(path) == 0:
                yield
                continue 
            next_coord = path.popleft()+Vec2d(0.5, 0.5)
            yield

           # turn()
            angle_diff = angle_between_vectors(self.tank.body.position,\
                                               next_coord)
            angle_next = periodic_difference_of_angles(\
                                self.tank.body.angle,angle_diff)
            self.tank.stop_moving()
            self.tank.body.angle = self.tank.body.angle % (2*math.pi)
            # Must check when tank is tilted downwords!
            if (angle_next < -math.pi) or (math.pi > angle_next > 0): 
                self.tank.turn_left()
            else:
                self.tank.turn_right()
                
            # correct_angle()
            while abs(angle_next) > MIN_ANGLE_DIF:
                angle_next = periodic_difference_of_angles(\
                                        self.tank.body.angle,angle_diff)
                yield
            self.tank.stop_turning()
            self.tank.accelerate()
            yield

            # correct_pos()
            current_dist = self.tank.body.position.get_distance(next_coord)
            temp_next_dist = current_dist
            while (current_dist >= temp_next_dist):
                self.update_grid_pos()
                # set new next tile when we reach the target!
                temp_next_dist = current_dist + 1 
                yield
            move_cycle = self.move_cycle_gen()

    def get_target_tile(self):
        """ Returns position of the flag if we don't have it. If we do
            have the flag, return the position of our home base. """
        if self.tank.flag is not None:
            x, y = self.tank.start_position
        else:
            self.get_flag() # Ensure that we have initialized it.
            x, y = self.flag.x, self.flag.y
        return Vec2d(int(x), int(y))

    def get_flag(self):
        """ This has to be called to get the flag, since we don't know
            where it is when the Ai object is initialized. """
        if self.flag is None:
        # Find the flag in the game objects list
            for obj in self.game_objects_list:
                if isinstance(obj, gameobjects.Flag):
                    self.flag = obj
                    break
        return self.flag

    def get_tile_of_position(self, position_vector):
        """ Converts and returns the float position of our tank to
            an integer position."""
        
        x, y = position_vector
        return Vec2d(int(x), int(y))

    def get_tile_neighbors(self, coord_vec):
        """ Returns all bordering grid squares of the input coordinate.
            A bordering square is only considered accessible if it is grass
            or a wooden box. """
        
        curr_tile = self.get_tile_of_position(coord_vec)
        tile1 = Vec2d(curr_tile[0]+1, curr_tile[1])
        tile2 = Vec2d(curr_tile[0]-1, curr_tile[1])
        tile3 = Vec2d(curr_tile[0], curr_tile[1]+1)
        tile4 = Vec2d(curr_tile[0], curr_tile[1]-1)
        neighbors = [tile1, tile2, tile3, tile4]        
        return filter(self.filter_tile_neighbors, neighbors)
      

    def filter_tile_neighbors (self, coord):
        """ Checks wether a tank can move to the specific tile type"""
        
        x = coord[0]
        y = coord[1]
        if (x >= 0 and x <= self.MAX_X) and (y >= 0 and y <= self.MAX_Y):
            box = self.currentmap.boxAt(x,y)
            if box == 0 or box == 2 or self.metal_path == True and box == 3:
                return True
        return False



